#!/usr/bin/env python3

import sys
import matplotlib
matplotlib.use("Agg")
print('[A] Loading ' + sys.argv[0] + '.\n')
import os # for reading files
import networkx as nx # for visualisation
from copy import deepcopy
from nltk.corpus import stopwords
import numpy as np # for calculations
import re
import spacy # for nlp
from multiprocessing import Pool
import random
import matplotlib.pyplot as plt
import config

nlp = spacy.load('en') # standard english nlp

def frequencies(target_string, search_result_list):
    """Counts occurrences of nodes and cooccurrences.
    
    Iterates over the corpus (and snippets provided with the task) line by line 
    and counts every token and tuple of tokens within a line (context). These 
    tokens is filtered by stop words, pos tags and context length.
    
    Args:
        target_string: contexts are selected if they contain this string. For
            further processing this string is removed from the contexts.
        search_result_list: List of titles and snippets provided with the task.
        
    Returns:
        node_freq_dict: Dictionary of occurrences of every eligible token
            within every context the target occurs in.
        edge_freq_dict: Dictionary of occurrences of every eligible tuple of
            tokens within every context the target occurs in.
    
    """

    corpus_path = config.corpus
    max_node_count = config.max_nodes
    max_edge_count = config.max_edges
    
    bracketed_target_string = '('+target_string+')'
    
    # Remove unnecessary tokens from snippets.
    _search_result_list = list()
    for r in search_result_list:
        r = r.replace('<b>', '')
        r = r.replace('</b>', '')
        r = r.replace(r'\\', '')
        r = r.strip()
        _search_result_list.append(r)
    
    # Initialise frequencies with counts from results.
    node_freq_dict, edge_freq_dict = process_file(_search_result_list,
                                                  target_string,
                                                  dict(),
                                                  dict()) 
    
    corpus_file_path_list = [corpus_path + f for f in os.listdir(corpus_path)]
    corpus_size = len(corpus_file_path_list)
    
    processed_file_count = 0
    for corpus_file_path in corpus_file_path_list:
        
        node_count = len(node_freq_dict)
        edge_count = len(edge_freq_dict)
        
        # Print update after every 11th of the corpus is parsed.
        if processed_file_count % int(corpus_size/11) == 0: 
            
            file_ratio = processed_file_count / corpus_size
            max_node_ratio = node_count / max_node_count
            max_edge_ratio = edge_count / max_edge_count
            
            ratios = [file_ratio, max_node_ratio, max_edge_ratio]
            
            # Use ratio closest to 100%.
            highest_ratio = int((max(ratios))*100)
            
            print('[a] ~{:02d}%\tNodes: {}\tEdges: {}\t{}.'.format(highest_ratio,
                                                                   node_count,
                                                                   edge_count,
                                                                   bracketed_target_string))
        
        if node_count > max_node_count:
            print('[a] 100%\tNodes: {}\tEdges: {}\t{}.'.format(node_count,
                                                               edge_count,
                                                               bracketed_target_string))
            return node_freq_dict, edge_freq_dict
        
        if edge_count > max_edge_count:
            print('[a] 100%\tNodes: {}\tEdges: {}\t{}.'.format(node_count,
                                                               edge_count,
                                                               bracketed_target_string))
            return node_freq_dict, edge_freq_dict
        
        with open(corpus_file_path, 'r') as corpus_file:
            
            node_freq_dict, edge_freq_dict = process_file(corpus_file,
                                                          target_string,
                                                          node_freq_dict,
                                                          edge_freq_dict)
        
        processed_file_count += 1
    
    print('[a] 100%\tNodes: {}\tEdges: {}\t{}.'.format(node_count,
                                                       edge_count,
                                                       bracketed_target_string))
    
    return node_freq_dict, edge_freq_dict


def process_file(context_list, target_string, node_freq_dict, edge_freq_dict):
    """Updates the counts of nodes and edges for a given document and target.
    
    Ammends the input dictionaries with counts from each context withing the
    list of contexts. Furthermore filters out small contexts and tokens from
    the stopword list or with wrong pos tags.
    
    Args:
        context_list: List of contexts (lines, paragraphs) that are to be
            considered for updating the counting dictionaries.
        target_string: Target string for filtering out every context that does 
            not contain it.
        node_freq_dict: Dictionary of occurrences of every eligible token
            within every context the target occurs in.
        edge_freq_dict: Dictionary of occurrences of every eligible tuple of
            tokens within every context the target occurs in.
    
    Returns:
        node_freq_dict: Updated version of the input node dict.
        edge_freq_dict: Updated version of the input edge dict.
    """
    
    spaced_target_string = target_string.replace('_', ' ')
    
    stopword_list = set(stopwords.words('english') + config.stop_words)
    allowed_tag_list = config.allowed_tags
    min_context_size = config.min_context_size
        
    try:
        
        for context in context_list:
            
            context = context.lower()
            if spaced_target_string in context: # Pre-select lines greedy.
                
                token_set = set()
                
                # Allow target to be treated as single entity.
                context = context.replace(spaced_target_string, target_string)
                processed_context = nlp(context)
                
                if target_string in [token.text for token in processed_context]:
                    
                    for token in processed_context:
                        
                        # Do not add target word to nodes.
                        if token.text == target_string:
                            pass
                        
                        # Do not add stop words to nodes.
                        elif token.text in stopword_list:
                            pass
                        
                        # Add only tokens with allowed tags to nodes.
                        elif token.tag_ in allowed_tag_list:
                            token_set.add(token.lemma_)
                            
                    context_size = len(token_set)
                    
                    if context_size >= min_context_size:
                        for token in token_set:
                            
                            if token in node_freq_dict:
                                node_freq_dict[token] += 1
                            else:
                                node_freq_dict[token] = 1
                        
                        #set of possible edges
                        for edge in {(x,y) for x in token_set for y in token_set if x < y}:
                            
                            if edge in edge_freq_dict:
                                edge_freq_dict[edge] += 1
                            else:
                                edge_freq_dict[edge] = 1
    
    # If file is corrupted (can't always be catched with if-else), ignore file.
    except UnicodeDecodeError:
        
        pass            
    
    return node_freq_dict, edge_freq_dict


def build_graph(node_freq_dict, edge_freq_dict):
    """Builds undirected weighted graph from dictionaries.
    
    Creates graph and appends every edge and node in the parameter dictionaries,
    given they occur frequently enough. For every edge a weight is calculated.
    
    Args:
        node_freq_dict: Dictionary of occurrences of every eligible token
            within every context the target occurs in.
        edge_freq_dict: Dictionary of occurrences of every eligible tuple of
            tokens within every context the target occurs in.
    
    Returns:
        cooccurrence_graph: Filtered undirected dice weighted small word 
            cooccurrence graph for a given target entity.
    """
    
    min_node_freq = config.min_node_freq
    min_edge_freq = config.min_edge_freq
    max_weight = config.max_weight
    
    cooccurrence_graph = nx.Graph()
    
    for node, frequency in node_freq_dict.items():
        
        if frequency >= min_node_freq:
            cooccurrence_graph.add_node(node)
            
    for node_tuple, frequency in edge_freq_dict.items():
        
        if frequency < min_edge_freq:
            
            continue
        
        elif node_tuple[0] not in cooccurrence_graph.nodes:
            
            continue
        
        elif node_tuple[1] not in cooccurrence_graph.nodes:
            
            continue
        
        else:
            
            cooccurrence_frequency = edge_freq_dict[node_tuple]
            node0_frequency = node_freq_dict[node_tuple[0]]
            node1_frequency = node_freq_dict[node_tuple[1]]
            
            prob_0 = cooccurrence_frequency / node0_frequency
            prob_1 = cooccurrence_frequency / node1_frequency
            
            best_weight = 1 - max(prob_0, prob_1)
            #dice_weight = 1 - ((prob_0 + prob_1) / 2)
            
            if best_weight <= max_weight:
                
                cooccurrence_graph.add_edge(*node_tuple, weight=best_weight)
            
            else:
                
                pass
    
    return cooccurrence_graph


def root_hubs(graph, edge_freq_dict):
    """Identifies senses (root hubs) by choosing nodes with high degrees
    
    Selects root hubs according to the algorithm in Véronis (2004). Nodes with
    high degree and neighbors with low weights (high cooccurrence) are chosen
    until there are no more viable candidates. A root hub candidate is every
    node that is not already a hub and is not a neighbor of one.
    
    Args:
        graph: Weighted undirected graph.
        edge_freq_dict: Dictionary of weights for every tuple in our graph.
        
    Returns:
        hub_list: List of root hubs, i.e. strings that are selected using the
            algorithm explained above.
    """
    
    min_neighbors = config.min_neighbors
    threshold = config.threshold
    
    # Allow operations on graph without altering original one.
    graph_copy = deepcopy(graph)
    
    # Sort according to degree (number of neighbors).
    candidate_list = sorted(graph_copy.nodes,
                            key=lambda node: graph_copy.degree[node],
                            reverse=True)
    
    hub_list = list()
    
    # While there are still candidates, search for root hubs.
    while candidate_list:
        
        candidate = candidate_list[0] #best hub candidate
        
        if graph_copy.degree[candidate] >= min_neighbors:
        
            by_frequency = lambda node: edge_freq_dict[candidate,node] \
                                         if candidate < node \
                                         else edge_freq_dict[node,candidate]
                              
            most_frequent_neighbor_list = sorted(graph_copy.adj[candidate],
                                                 key=by_frequency,
                                                 reverse=True) [:min_neighbors] 
            
            # If the mean weight of the most frequent neighbors cooccur
            # frequently enough with candidate, the candidate is approved.
            if np.mean([graph_copy.edges[candidate,node]['weight'] 
                        for node in most_frequent_neighbor_list])  < threshold: 
                
                # Add candidate as root hub.
                hub_list.append(candidate)
            
                # Remove neighbors of new hub as hub candidates.
                for neighbor in deepcopy(graph_copy).adj[candidate]:
                    graph_copy.remove_node(neighbor)
                
            # Remove hub candidate.
            graph_copy.remove_node(candidate)
            
            # Reorder potential hubs after deletions.
            candidate_list = sorted(graph_copy.nodes,
                                    key=lambda node: graph_copy.degree[node],
                                    reverse=True)
        
        else:
        
            return hub_list
    
    return hub_list


def components(graph, root_hub_list, target_string):
    """Builds minimum spanning tree from graph and removes singletons.
    
    Applies components algorithm from Véronis (2004) and removes singletons.
    
    Args:
        graph: Undirected weighted graph.
        root_hub_list: List of strings of root hubs of graph.
        target_string: Root of minimum spanning tree.
        
    Returns:
        minimum_spanning_tree: Minimum spanning tree with target as
            root and root hubs as direct children. Singletons removed.
    """
    
    graph_copy = deepcopy(graph)
    
    graph_copy.add_node(target_string)
    for root_hub in root_hub_list:
        graph_copy.add_edge(target_string,root_hub,weight=0)
        
    minimum_spanning_tree = nx.minimum_spanning_tree(graph_copy)
    
    # Remove singletons, deepcopy for iteration while being altered.
    for node in deepcopy(minimum_spanning_tree).nodes:
        if len(minimum_spanning_tree.adj[node]) == 0:
            minimum_spanning_tree.remove_node(node)
    
    return minimum_spanning_tree


def score(graph, component, root_hub_list):
    """Calculate score for a given component in a minimum spanning tree.
    
    First the correct root for the component is chosen. If no root hub is
    suitable, an empty array is returned. A score is calculated for the distance
    of the component and its root and returned as part of an array filled with
    zeroes.
    
    Args:
        graph: Minimum spanning tree.
        component: Node (string) from which the distances are to be calculated.
        root_hub_list: List of strings of root hubs (senses) of original graph.
    
    Returns:
        score_array: Array with one score for the correct root hub and filled 
            with zeroes..
    """
    
    root_hub_count = len(root_hub_list)
    
    #Initialise score array.
    score_array = np.zeros(root_hub_count)
    
    # Find root of component.
    distance_list = list()
    for root_hub in root_hub_list:
        if nx.has_path(graph, component, root_hub):
            distance_list.append(1/(1+len(nx.shortest_path(graph, component, root_hub))))
        else:
            distance_list.append(0)
            
    if sum(distance_list) == 0:
        return score_array
            
    root_idx = np.argmax(distance_list)
    root = root_hub_list[root_idx]
    
    shortest_path = nx.shortest_path(graph, component, root, 'weight')
    total_weight = 0

    # Add weights of every sub-path.
    for i in range(1, len(shortest_path)):
        sub_from, sub_to = shortest_path[i-1], shortest_path[i]
        total_weight += graph[sub_from][sub_to]['weight']

    score_array = np.zeros(root_hub_count)
    score_array[root_idx] = 1/(1+total_weight)
    
    return score_array


def induce(topic_name, result_list):
    """


    """
    
    statistics = dict()
            
    #removes trailing new_lines
    old_target_string = topic_name.strip() #original target
    
    if old_target_string.strip() in [f.replace('.absinth', '') for f in os.listdir(config.output)]:
        return None
    
    statistics['target'] = old_target_string
    
    #in topics longer than two words, the leading 'the' can generally be removed without changing the sense
    if old_target_string[:4] == 'the_' and old_target_string.count('_') >= 2:
        
        target_string = old_target_string[4:]
        
    else:
        
        target_string = old_target_string
    
    #counts occurences of single words, as well as cooccurrences, saves it in dictionary
    print('[a]', 'Counting nodes and edges.\t('+old_target_string+')')
    node_freq_dict, edge_freq_dict = frequencies(target_string, result_list[topic_id])
    
    #builds graph from these dictionaries, also applies multiple filters
    print('[a]', 'Building graph.\t('+old_target_string+')')
    G = build_graph(node_freq_dict, edge_freq_dict)
    
    statistics['node count'] = len(G.nodes)
    statistics['edge count'] = len(G.edges)

    #finds root hubs (senses) within the graph + more filters for these
    print('[a]', 'Collecting root hubs.\t('+old_target_string+')')
    H = root_hubs(G, edge_freq_dict)
    
    #adds sense inventory to buffer with some common neighbors for context
    statistics['hubs'] = dict()
    for h in H:
        mfn = sorted(G.adj[h], key=lambda x: edge_freq_dict[h,x] if h < x else edge_freq_dict[x, h], reverse=True)[:6]
        statistics['hubs'][h] = mfn
    
    #performs minimum_spanning_tree algorithm on graph
    print('[a]', 'Building minimum spanning tree.\t('+old_target_string+')')
    T = components(G, H, target_string)

    return T, H, statistics


def disambiguate(minimum_spanning_tree, root_hub_list,
                 context_list, target_string):
    """Matches contexts to senses.
    
    Adds up scores for each token in a context string and matches the context
    to the root hub with the highest score.
    
    Args:
        minimum_spanning_tree: Minimum spanning tree with target as root.
        root_hub_list: List of strings of root hubs (senses).
        context_list: List of sentence strings that are to be clustered.
        target_string: String of target word, also root of MST.
    
    Returns:
        mapping_dict: Dictionary of root hubs (senses) as keys and context ids
            as values.
    """
    
    target_string = target_string.replace('_', ' ')
    context_list = [context.lower().strip().replace(target_string, '')
                    for context in context_list]
    
    score_dict = dict() #memoisation for scores
    mapping_dict = {topic:[] for topic in range(1,len(root_hub_list)+1)}
    
    #if no sense is found for a target word, we should assume that there only is one sense
    if len(root_hub_list) == 0: 
            
        return {0:[i for i in range(1, len(context_list)+1)]}
    
    idx = 0
    
    for context in context_list:
        
        idx += 1 #index based on position in list
    
        processed_context = nlp(context)
        text_list = [token.text for token in processed_context] #tokens
        
        score_array = np.zeros(len(root_hub_list)) #initialise with zeros for every sense
        
        for text in text_list:
            
            if text in minimum_spanning_tree.nodes: #if word wasn't filtered out
                
                if text in score_dict: #memoisation
                        
                    new_scores = score_dict[text]
                
                else:
                    
                    new_score = score(minimum_spanning_tree, 
                                      text, root_hub_list)
                    score_dict[text] = new_score #memoisation
                    
                score_array += new_score
            
            else:
            
                pass
        
        # If disambiguator does not detect a sense, return singleton.
        if np.max(score_array) == 0:
            
            pass
            
        else:
                
            # Apply sense with the highest score to context
            max_score = np.max(score_array)
            argmax_score = np.argmax(score_array)
            
            # Clusters begin at 1
            mapping_dict[argmax_score + 1].append(idx)

    return mapping_dict


def main(topic_id, topic_name, result_list):
    """


    """
    
    print('[a]', 'Inducing word senses for {}.'.format(topic_name))
    T, H, statistics = induce(topic_name, result_list)
    
    #matches senses to clusters
    print('[a]', 'Disambiguating result_list.\t('+old_target_string+')')
    D = disambiguate(T, H, result_list[topic_id], target_string)
    
    #collect statistics from result.
    cluster_count = 0
    cluster_length_list = list()
    for cluster,result_list in D.items():
        cluster_length = len(result_list)
        if cluster_length != 0:
            cluster_count += 1
            cluster_length_list.append(cluster_length)
    statistics['mean_cluster_length'] = np.mean(cluster_length_list)
    statistics['cluster_count'] = cluster_count

    #prints buffer
    print('[a]', 'Writing to file.\t('+old_target_string+')')
    

    f = open(output_path+old_target_string+'.absinth', 'w')

    f.write('subTopicID\tresultID\n')

    #writes clustering to file
    for cluster,result_list in D.items():
        for result in result_list:
            f.write(topic_id+'.'+str(cluster)+'\t'+topic_id+'.'+str(result)+'\n')
        
    f.close()
        

def read_dataset(data_path):
    
    # results.txt includes the queries for a given target word
    results = dict()
    
    with open(data_path+'results.txt', 'r') as results_file:
        
        for line in results_file.readlines()[1:]:
            
            l = line.split('\t')
            id1, _ = l[0].split('.') #the second part of the id is ignored, as it is identical to the list index
            
            if id1 not in results:
                results[id1]=list()
                
            results[id1].append(" ".join(l[2:])) # here I join title and snippet, the URL is ignored
            
    
    # topics.txt is a list of target words
    topics = dict()
    
    with open(data_path+'topics.txt', 'r') as topics_file:
        
        for line in topics_file.readlines()[1:]:
            
            l = line.split('\t')
            topics[l[0]] = l[1]
    
    return results, topics




if __name__ == '__main__':
    # If absinth.py is run in test environment.
    if '-t' in sys.argv:
        data_path = config.test
    else:
        data_path = config.dataset
        
    results, topics = read_dataset(data_path)
    
    # Enables manual setting of process count.
    if '-p' in sys.argv:
        process_count = int(sys.argv[sys.argv.index('-p') + 1])
    else:
        process_count = 1
    
    with Pool(process_count) as pool:
        parameter_list = [(topic_id, topic_name, results)
                          for topic_id,topic_name in topics.items()]
        pool.starmap(main, parameter_list)
        
    #for topic_id,topic_name in topics.items():
       #word_sense_induction(topic_id,topic_name, results)
